/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: ClientHull.cpp 2194 2017-04-02 01:51:40Z stephane $
 * See license.txt.
 */

#include "ClientHull.hpp"
#include "TCPConnection.hpp"
#include "logging.hpp"
#include "tun_tap.hpp"
#include "Tracker.hpp"
#include "Packet.hpp"
#include <boost/asio.hpp>
#include <boost/chrono/system_clocks.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/regex.hpp>
#ifndef EPS_WINDOWS
	#include <unistd.h>
#endif
#include <thread>


#ifndef INVALID_HANDLE_VALUE
#define INVALID_HANDLE_VALUE -1
#endif


EPS::ClientHull::ClientHull(int argc, char *argv[]) :
	Hull(EPS::kClient, argc, argv),
	device_fd(INVALID_HANDLE_VALUE),
	tun_tap_stream(io_service)
{
	LOG_DEBUG << "outer client hull constructed";

	return;
}


EPS::ClientHull::~ClientHull(void)
{
	LOG_DEBUG << "outer client hull destruction in progress";
	this->done = true;

	if (device_fd != INVALID_HANDLE_VALUE)
	{
		LOG_DEBUG << "closing TUN/TAP interface " << device_name;
		EPS::close_tun_tap_interface(device_name, device_fd);

		#ifdef BOOST_ASIO_HAS_POSIX_STREAM_DESCRIPTOR
			close(device_fd);
		#else
			CloseHandle(device_fd);
		#endif
		device_fd = INVALID_HANDLE_VALUE;
	}

	// only re-write the hosts file if we have a change to save
	const std::string hosts1 = EPS::read_file("/etc/hosts");
	const std::string hosts2 = this->clean_hosts_file(hosts1);
	if (hosts1 != hosts2)
	{
		EPS::write_file("/etc/hosts", hosts2);
	}

	return;
}


void EPS::ClientHull::run_second_stage(void)
{
	while ( ! done )
	{
		auto until = std::chrono::system_clock::now() + seconds_between_retries;

		this->async_find_and_connect_to_server();
		this->setup_listener();
		this->start_connect_timer();

		if (not this->ec)
		{
			LOG_INFO << "running IO loop";
			this->io_service.run(this->ec);
		}

		if (this->ec)
		{
			LOG_ERROR << "EPS client has failed: " << this->ec.message();
			this->ec.clear();
		}
		if ( ! done )
		{
			// prevent busy loops
			std::this_thread::sleep_until(until);
		}
	}

	return;
}


void EPS::ClientHull::async_find_and_connect_to_server(void)
{
	// find all of the space or comma separated server names and IP addresses, look up each one, and attempt to connect to it
	std::string servers = this->addr;
	LOG_DEBUG << "server field is set to \"" << servers << "\"";

	// keep track of which server IP addresses to which we've connected (so we don't connect twice to one address)
	static EPS::IPv4Set server_ip_addresses;
	server_ip_addresses.clear();

	while (not servers.empty())
	{
		const size_t pos = servers.find_first_of(" ,;");

		std::string next_server;
		if (pos == std::string::npos)
		{
			// only one server left to read
			next_server.swap(servers);
		}
		else
		{
			next_server = servers.substr( 0, pos );
			servers.erase( 0, pos + 1 );
		}

		if (not trim(next_server).empty())
		{
			LOG_INFO << "next server to which we'll try to connect:  \"" << next_server << "\"";

			// look up the server's address(es)
			const boost::asio::ip::tcp::resolver::query::flags flags(	boost::asio::ip::tcp::resolver::query::address_configured	|
																		boost::asio::ip::tcp::resolver::query::numeric_service		);
			const boost::asio::ip::tcp::resolver::query query(next_server, std::to_string(this->port), flags);
			this->tcp_resolver.async_resolve( query,
					[=](const boost::system::error_code &error, boost::asio::ip::tcp::resolver::iterator iter)
					{
						const boost::asio::ip::tcp::resolver::iterator end;
						if (ec || iter == end)
						{
							LOG_ERROR << "failed to resolve server name \"" + next_server + ":" + std::to_string(this->port) + "\": ec=" + error.message();
						}
						else
						{
							// create a new async connection for each of the addresses returned by async_resolve()
							for (; iter != end; iter ++ )
							{
								const EPS::TCPEP tcp_endpoint = iter->endpoint();
								const EPS::IPv4 ip = tcp_endpoint.address().to_v4();

								if (server_ip_addresses.count(ip) > 0)
								{
									LOG_INFO << "ignoring duplicate server address \"" << ip.to_string() << "\" to which we've already opened a beam";
								}
								else
								{
									// remember this address so we don't try to connect to the same place multiple times
									server_ip_addresses.insert(ip);

									LOG_INFO << "creating a new tcp beam to the server at " << tcp_endpoint;
									EPS::TConnPtr ptr(new EPS::TCPConnection(tcp_endpoint));
									this->server_conns.insert(ptr);
									ptr->async_connect();
								}
							}
						}
					} );
		}
	}

	return;
}


void EPS::ClientHull::connection_failed( const EPS::TConnPtr &connection )
{
	/* It is important to know that this function can be called on a
	 * connection that has not yet been inserted into the tracker.
	 */

	if (connection)
	{
		LOG_WARN << connection->name << "connection has failed and will be deleted";

		connection->socket.cancel(this->ec);
		this->stats.minute_stats.connections_dropped ++;
		this->server_conns.erase(connection);

		if (this->tracker.exists(connection))
		{
			EPS::TrackerData &data = this->tracker.get(connection);
			if (data.connection == connection)
			{
				LOG_WARN << connection->name << "resetting active direct connection";
				data.connection.reset();
			}
			if (data.pending_mesh_connection == connection)
			{
				LOG_WARN << connection->name << "resetting pending direct connection";
				data.pending_mesh_connection.reset();
			}
		}
	}

	if (this->server_conns.empty())
	{
		THROW_RUNTIME_ERROR(-1, "No connection remains to the EPS server.");
	}

	return;
}


void EPS::ClientHull::processComm( const EPS::TConnPtr &connection, EPS::NetBuffer *nb, const EPS::CommHeader1 *header, const EPS::ECommType type )
{
	// By the time this function is called, a full packet has been read into connection->incoming_array, and then decrypted into connection->crypt_buffer.

	switch (type)
	{
		case EPS::kData1:
		{
			// push this packet back out the tap interface (do not re-encrypt!)

			const EPS::CommData1 *	cd1						= reinterpret_cast<const EPS::CommData1*>(header);
			const size_t			number_of_bytes			= cd1->header.size_of_packet();
			const size_t			number_of_data_bytes	= number_of_bytes - sizeof(EPS::CommData1);

//			LOG_TRACE << connection->name << "injecting " << number_of_data_bytes << "-byte " << EPS::proto_to_string(cd1) << " packet into the local network";

			const boost::asio::const_buffers_1 buffer(cd1->data, number_of_data_bytes);
			const size_t bytes_written = boost::asio::write(this->tun_tap_stream, buffer, ec);

			stats.minute_stats.tap_bytes_sent		+= bytes_written;
			stats.minute_stats.tap_packets_sent		++;

			break;
		}
		case EPS::kList2:
		{
			if (this->server_conns.count(connection) == 0)
			{
				THROW_RUNTIME_ERROR(-1, "received peer list from non-server connection");
			}

			const EPS::CommList2		*	cl2		= reinterpret_cast<const EPS::CommList2*>(header);
			const EPS::CommList2Data	*	data2	= cl2->data;
			const size_t					entries	= ntohs(cl2->number_of_records);

			LOG_INFO << "received a new list of peers: number of entries: " << entries;

			for (size_t idx = 0; idx < entries; idx++)
			{
				EPS::IPv4Set other_addresses;
				for (size_t addr_idx = 0; addr_idx < MAX_NUMBER_OF_OTHER_ADDRESSES; addr_idx ++)
				{
					const EPS::IPv4 ip( ntohl(data2[idx].other_addresses[addr_idx]) );
					if (ip.is_unspecified	() == false	&&
						ip.is_loopback		() == false	)
					{
						other_addresses.insert(ip);
					}
				}

				const EPS::IPv4 address_eps( ntohl(data2[idx].address_eps) );
				EPS::UUID uuid;
				memcpy(uuid.data, data2[idx].uuid, uuid.static_size());
				const std::string hostname( data2[idx].hostname_str );

				std::string append;
				if (uuid == this->uuid)
				{
					append = "  <-- THIS IS US";
				}
				LOG_INFO << "list #" << idx << ": " << uuid << ": " << ipset_to_string(other_addresses) << " aka " << address_eps << " [" << hostname << "]" << append;

				if (this->tracker.exists(uuid))
				{
					EPS::TrackerData &data = this->tracker.update_entry(uuid);
					data.hostname = hostname;
					if (data.address_eps		!= address_eps		||
						data.other_addresses	!= other_addresses	)
					{
						// client has changed IP address, so drop whatever old connection we thought we had
						if (data.is_active())
						{
							LOG_WARN << data.connection->name << "addresses have changed, connection will be dropped";
							this->connection_failed(data.connection);
						}
						data.address_eps		= address_eps;
						data.other_addresses	= other_addresses;
						LOG_INFO << "updated addresses: " << data.as_debug();
					}
				}
				else
				{
					EPS::TrackerData &data = this->tracker.add_inactive_entry(uuid, address_eps);
					data.other_addresses	= other_addresses;
					data.mode				= EPS::kClient;
					data.hostname			= hostname;
					LOG_TRACE << "added a new entry for " << data.as_debug();
				}
			}

			this->tracker.deactivate_old_entries();

			this->update_hosts_file();

			break;
		}
		case EPS::kAddress1:
		{
			if (this->server_conns.count(connection) == 0)
			{
				THROW_RUNTIME_ERROR(-1, "received EPS address requests from non-server connection");
			}

			const EPS::CommAddress1 *ca1 = reinterpret_cast<const EPS::CommAddress1*>(header);
			this->network = IPv4(ntohl(ca1->network_address	));
			this->netmask = IPv4(ntohl(ca1->network_netmask	));
			this->address = IPv4(ntohl(ca1->client_address	));
			LOG_INFO << connection->name << "address: " << this->address.to_string(ec) << ", network: " << this->network.to_string(ec) << ", netmask: " << this->netmask.to_string(ec);

			if (EPS::validate_network_against_existing_routes(this->network.to_ulong(), this->netmask.to_ulong()) == false)
			{
				THROW_RUNTIME_ERROR(-1, "EPS network address conflicts with existing network or route");
			}
			EPS::raise_tun_tap_interface(ca1->client_address, this->device_name, this->device_fd);
			EPS::set_tun_tap_routes(ca1->network_address, ca1->network_netmask, this->device_name );
			this->tun_tap_stream.assign(this->device_fd, this->ec);
			if (ec)
			{
				THROW_RUNTIME_ERROR(errno, "Failed to create the TUN/TAP stream: " + ec.message());
			}

			this->async_read_tun_tap();

			break;
		}
		default:
		{
			break;
		}
	}

	return;
}


std::string EPS::ClientHull::clean_hosts_file(std::string hosts)
{
	const std::string signature = "EPS HOST ENTRIES FOR " + this->address.to_string();

	if (hosts.empty())
	{
		hosts = EPS::read_file("/etc/hosts");
	}

	const boost::regex rx(
		"^"							// start of buffer
		"(.*?)"						// non-greedy "match anything" (non-greedy so we don't consume a bunch of newlines)
		"\\s*"						// whitespace
		"# START "	+ signature +	// start of EPS signature
		".*"						// EPS entries
		"# STOP "	+ signature +	// end of EPS signature
		"\\s*"						// whitespace
		"(.*)"						// anything else
		"$"							// end of buffer
		);

	boost::smatch what;
	if (boost::regex_match(hosts, what, rx))
	{
		// exclude the part we want to ignore and keep just the other two parts
		std::stringstream ss;
		ss	<< what[1].str() << std::endl << what[2].str();
		hosts = ss.str();
	}

	return hosts;
}


void EPS::ClientHull::update_hosts_file(void)
{
	const std::string signature = "EPS HOST ENTRIES FOR " + this->address.to_string();

	const EPS::TrackerMap &tm = this->tracker.get_all_entries();

	EPS::StrMap m;
	for (auto iter : tm)
	{
		const EPS::TrackerData &td = iter.second;
		if (td.is_client()							&&
			td.hostname.empty()			== false	&&
			td.other_addresses.empty()	== false	&&
			td.uuid != this->uuid					)
		{
			LOG_TRACE << "need to create a hosts entry for " << td.as_debug();
			const std::string eps_address = td.address_eps.to_string();
			if (m.count(eps_address))
			{
				LOG_WARN << "address " << eps_address << " is ambiguous: " << m[eps_address] << " or " << td.hostname << "?";
			}

			m[eps_address] = td.hostname;
		}
	}

	const std::string original_hosts = EPS::read_file("/etc/hosts");
	std::string hosts = this->clean_hosts_file(original_hosts);
	if (not m.empty())
	{
		std::stringstream ss;
		ss	<< std::endl
			<< std::endl
			<< "# START " << signature << std::endl
			<< "# Number of host entry lines managed automatically by EPS Conduits v" << EPS_VERSION << ": " << m.size() << std::endl;

		for (auto iter : m)
		{
			ss << iter.first << " " << iter.second << std::endl;
		}
		ss << "# STOP " << signature << std::endl << std::endl;

		hosts += ss.str();
	}

	if (original_hosts != hosts)
	{
		EPS::write_file("/etc/hosts", hosts);
	}
	else
	{
		LOG_TRACE << "no changes made to /etc/hosts";
	}

	return;
}


void EPS::ClientHull::async_read_tun_tap(void)
{
	const size_t minimum_number_of_bytes_expected = EPS::cipher_block_size; // seems like a good size...?

	EPS::NetBuffer *nb = new EPS::NetBuffer(EPS::NetBuffer::EType::kFullyDecrypted);

	// special case:  we need to leave some room for the CommData1 header prior to the data bytes
	const size_t offset = sizeof(EPS::CommData1);
	boost::asio::mutable_buffers_1 buffer(nb->buffer.data() + offset, nb->buffer.max_size() - offset);
	boost::asio::async_read(this->tun_tap_stream, buffer, boost::asio::transfer_at_least(minimum_number_of_bytes_expected),
			[=](const EC &error, std::size_t bytes_read)
			{
				static int milliseconds_to_sleep = 0;

				if (error || bytes_read < minimum_number_of_bytes_expected)
				{
					if (error == boost::asio::error::eof)
					{
						/** @todo See this regression in kernel 3.19 and above:
						 * https://bugzilla.kernel.org/show_bug.cgi?id=96381
						 * The problem is the tun/tap API is incorrectly returning 0 bytes instead of EAGAIN when
						 * nothing is available to read.  To work around this, let's insert increasingly long pauses
						 * if the error is EOF.
						 */
						milliseconds_to_sleep ++;
						if (milliseconds_to_sleep > 10)
						{
							if (milliseconds_to_sleep > 100)
							{
								milliseconds_to_sleep = 100;
							}
							std::this_thread::sleep_for( std::chrono::milliseconds(milliseconds_to_sleep) );
						}
					}
					else
					{
						LOG_ERROR << "failed to read from tun/tap device (" << bytes_read << " bytes, error flag is " << (error?"":"not ") << "set): " << error.message();
						std::this_thread::sleep_for( std::chrono::milliseconds(100) );	// prevent a tight loop if something has gone wrong
					}
					delete nb;
				}
				else
				{
					// otherwise we need to process this tun/tap packet
					milliseconds_to_sleep = 0;

//					LOG_TRACE << "finished reading " << bytes_read << " bytes from TUN/TAP device";
					this->stats.minute_stats.tap_bytes_received		+= bytes_read;
					this->stats.minute_stats.tap_packets_received	++;
					// beware, the "length" is tricky here since the starting offset is not zero!
					nb->length = bytes_read;
					this->process_tun_tap_packet(nb);
				}

				if (not this->done)
				{
					// schedule the next async read from the tun/tap device
					this->async_read_tun_tap();
				}
			} );

	return;
}


void EPS::ClientHull::process_tun_tap_packet(EPS::NetBuffer *nb)
{
	/** Called once a packet/frame has been copied into the specified NetBuffer.
	 *
	 *	- byte[0-1]:	flags (network byte order)
	 *	- byte[2-3]:	proto (network byte order)
	 *	- byte[4+...]:	ethernet frame
	 *	- byte[...]:	ip packet
	 *	...etc...
	 *
	 * Note that if IFF_NO_PI is specified when the TAP interface is created (which is what %EPS does since v0.0.1-549)
	 * then the flags and proto bytes aren't included, and the read starts directly at the ethernet frame:
	 *
	 *	- byte[0+.]:	ethernet frame
	 *	- byte[...]:	ip packet
	 *	...etc...
	 */

	bool send_via_tcp	= false;	// assume we can send via udp
	UUID dst_uuid		= boost::uuids::nil_uuid();

	EPS::CommData1 *cd1 = reinterpret_cast<EPS::CommData1*>(nb->data());

	// send this packet either to the eps server, or to the appropriate client if we have a direct connection
	EPS::TConnPtr connection_to_use = *this->server_conns.begin();

	// see if we can figure out who this packet is destined to
	EPS::Packet packet(cd1->data);
//	packet.debug();
	const EPS::Ipv4Header *ipv4_header = packet.get_ipv4_header();
	if (ipv4_header)
	{
		if (ipv4_header->protocol == IPPROTO_TCP)
		{
			// if this packet is TCP, then send using the TCP beam
			send_via_tcp = true;
		}

		const IPv4 dst(ntohl(ipv4_header->dst_ip));
		if (this->tracker.does_not_exist(dst))
		{
			static	size_t	packet_counter	= 0;
			static	time_t	timestamp		= 0;
			const	time_t	now				= std::time(nullptr);

			packet_counter ++;

			if (timestamp == 0)
			{
				timestamp = now;
			}
			else if (now > timestamp + 300 || packet_counter >= 1000)
			{
				LOG_WARN << packet_counter << " packets sent to unknown peers such as " << dst.to_string() << " in the last " << EPS::get_approximate_time(timestamp);
				timestamp		= 0;
				packet_counter	= 0;
			}
		}
		else
		{
			dst_uuid = this->tracker.uuid_for_destination(dst);

			const EPS::TrackerData &data = this->tracker.get(dst_uuid);
			if (data.connection)
			{
				// may as well use the direct connection we have!
				connection_to_use = data.connection;
			}
			else
			{
				// ...would be nice if we had a working connection to this peer
				if (this->pending_uuids.count(dst_uuid) == 0)
				{
					this->pending_uuids.insert(dst_uuid);
					LOG_TRACE << "remembering peer for a direct mesh connection: " << data.as_debug();
				}
				else
				{
				}
			}
		}
	}

	cd1->set(nb->length, dst_uuid);
	nb->length = cd1->header.size_of_packet();
//	LOG_TRACE << "data1 packet from tun_tap interface: " << EPS::hex_out(nb->data(), nb->length);

	if (not send_via_tcp)
	{
		/// @todo check to see if the UDP channel is up
		// for now, until udp is working, always send via tcp
		send_via_tcp = true;
	}

	if (send_via_tcp)
	{
		this->async_write(connection_to_use, nb);
	}
	else
	{
		/// @todo send this packet over UDP instead of TCP
		THROW_RUNTIME_ERROR(-1, "udp channel not yet supported");
	}

//	LOG_TRACE << connection_to_use->name << "sent " << bytes_sent << " bytes to the EPS server";

	return;
}
