/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Cipher.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#include "Cipher.hpp"
#include "logging.hpp"
#include <fstream>
#include <openssl/rand.h>
#include <openssl/sha.h>


std::array<uint8_t, 32>	EPS::Cipher::key;					// 256-bit
std::array<uint8_t, 16>	EPS::Cipher::iv;					// 128-bit
const EVP_CIPHER	*	EPS::Cipher::cipher					= nullptr;
EVP_CIPHER_CTX		*	EPS::Cipher::decrypt_ctx			= nullptr;
EVP_CIPHER_CTX		*	EPS::Cipher::encrypt_ctx			= nullptr;
std::string				EPS::Cipher::hash					= "<invalid>";
std::string				EPS::Cipher::name					= "<invalid>";
size_t					EPS::Cipher::size_of_keys_in_bytes	= 0;
size_t					EPS::Cipher::size_of_iv_in_bytes	= 0;


void EPS::Cipher::reset(void)
{
	LOG_TRACE << "taking down encryption/decryption ciphers";

	if (EPS::Cipher::decrypt_ctx)
	{
		EVP_CIPHER_CTX_cleanup		(EPS::Cipher::decrypt_ctx);
		EVP_CIPHER_CTX_free			(EPS::Cipher::decrypt_ctx);
		EPS::Cipher::decrypt_ctx	= nullptr;
	}

	if (EPS::Cipher::encrypt_ctx)
	{
		EVP_CIPHER_CTX_cleanup		(EPS::Cipher::encrypt_ctx);
		EVP_CIPHER_CTX_free			(EPS::Cipher::encrypt_ctx);
		EPS::Cipher::encrypt_ctx	= nullptr;
	}

	return;
}


void EPS::Cipher::initialize(EPS::StrMap &configuration)
{
	EPS::Cipher::iv						.fill('\0');
	EPS::Cipher::key					.fill('\0');
	EPS::Cipher::cipher					= EVP_aes_256_cbc		(					);
	EPS::Cipher::size_of_keys_in_bytes	= EVP_CIPHER_key_length	(EPS::Cipher::cipher);
	EPS::Cipher::size_of_iv_in_bytes	= EVP_CIPHER_iv_length	(EPS::Cipher::cipher);
	EPS::Cipher::name					= EVP_CIPHER_name		(EPS::Cipher::cipher);

	if (EPS::Cipher::iv.max_size() != EPS::Cipher::size_of_iv_in_bytes)
	{
		LOG_ERROR << EPS::Cipher::name + " iv requires " << EPS::Cipher::size_of_iv_in_bytes * 8 << " bits, but buffer contains " << EPS::Cipher::iv.max_size() * 8 << " bits";
		THROW_RUNTIME_ERROR(-1, "cannot enable encryption/decryption due to incorrectly-sized IV cipher buffer");
	}

	if (EPS::Cipher::key.max_size() != EPS::Cipher::size_of_keys_in_bytes)
	{
		LOG_ERROR << EPS::Cipher::name + " key requires " << EPS::Cipher::size_of_keys_in_bytes * 8 << " bits, but buffer contains " << EPS::Cipher::key.max_size() * 8 << " bits";
		THROW_RUNTIME_ERROR(-1, "cannot enable encryption/decryption due to incorrectly-sized key cipher buffer");
	}

	EPS::Cipher::get_cipher_key_and_iv(configuration);

	/* I'd like the users to have a (relatively) quick way to determine if the cipher IV and KEY have been set
	 * correctly, so compute a simple numeric hash.  If the hash isn't the same on the server and the client,
	 * then it will hopefully give the users a hint as to why EPS isn't working correctly.
	 */
	uint8_t sha1_hash[SHA_DIGEST_LENGTH];
	SHA_CTX ctx;
	SHA1_Init(&ctx);
	SHA1_Update(&ctx, EPS::Cipher::key.data(), EPS::Cipher::key.max_size());
	SHA1_Update(&ctx, EPS::Cipher::iv .data(), EPS::Cipher::iv .max_size());
	SHA1_Final(sha1_hash, &ctx);

	// SHA1 is a 20-byte value, which rendered as a hex string gives us a 40 character string
	const std::string hash_str = EPS::to_hex_string(sha1_hash, SHA_DIGEST_LENGTH);

	// pick a few characters from the 40-byte long SHA hex string and use that as our short nonsense cipher hash
	EPS::Cipher::hash =
			hash_str.substr(0 , 3)	+ "-" +
			hash_str.substr(15, 3)	+ "-" +
			hash_str.substr(30, 3);
	LOG_INFO << "openssl encryption/decryption cipher hash: " << EPS::Cipher::hash;

	EPS::Cipher::decrypt_ctx		= EVP_CIPHER_CTX_new();
	EVP_CipherInit_ex				(EPS::Cipher::decrypt_ctx, cipher, nullptr, nullptr, nullptr, ENC_MODE_DECRYPT);
	EVP_CIPHER_CTX_set_key_length	(EPS::Cipher::decrypt_ctx, (int)EPS::Cipher::key.max_size());
	EVP_CIPHER_CTX_set_padding		(EPS::Cipher::decrypt_ctx, 0);

	EPS::Cipher::encrypt_ctx		= EVP_CIPHER_CTX_new();
	EVP_CipherInit_ex				(EPS::Cipher::encrypt_ctx, cipher, nullptr, nullptr, nullptr, ENC_MODE_ENCRYPT);
	EVP_CIPHER_CTX_set_key_length	(EPS::Cipher::encrypt_ctx, (int)EPS::Cipher::key.max_size());
	EVP_CIPHER_CTX_set_padding		(EPS::Cipher::encrypt_ctx, 0);

	LOG_INFO << "openssl encryption/decryption cipher text: " << EPS::Cipher::name << ", " << EPS::Cipher::size_of_keys_in_bytes << "-byte key, " << EPS::Cipher::size_of_iv_in_bytes << "-byte iv";

	return;
}


void EPS::Cipher::get_cipher_key_and_iv(EPS::StrMap &configuration)
{
	bool save_iv	= false;
	bool save_key	= false;

	if (configuration.count("cipher_iv"))
	{
		EPS::from_hex_string(EPS::Cipher::iv.data(), EPS::Cipher::iv.max_size(), configuration["cipher_iv"]);
		LOG_INFO << "imported existing " << EPS::Cipher::iv.max_size() * 8 << "-bit cipher iv";
	}
	else
	{
		RAND_bytes(EPS::Cipher::iv.data(), (int)EPS::Cipher::iv.max_size());
		configuration["cipher_iv"] = EPS::to_hex_string(EPS::Cipher::iv.data(), EPS::Cipher::iv.max_size());
		LOG_INFO << "generated a new " << EPS::Cipher::iv.max_size() * 8 << "-bit cipher iv";
		save_iv = true;
	}

	if (configuration.count("cipher_key"))
	{
		EPS::from_hex_string(EPS::Cipher::key.data(), EPS::Cipher::key.max_size(), configuration["cipher_key"]);
		LOG_INFO << "imported existing " << EPS::Cipher::key.max_size() * 8 << "-bit cipher key";
	}
	else
	{
		RAND_bytes(EPS::Cipher::key.data(), (int)EPS::Cipher::key.max_size());
		configuration["cipher_key"] = EPS::to_hex_string(EPS::Cipher::key.data(), EPS::Cipher::key.max_size());
		LOG_INFO << "generated a new " << EPS::Cipher::key.max_size() * 8 << "-bit cipher key";
		save_key = true;
	}

	if (save_iv || save_key)
	{
		std::ofstream conf_file(configuration["conf_filename"], std::ofstream::out | std::ofstream::app);
		if (not conf_file)
		{
			THROW_RUNTIME_ERROR(errno, "Cannot save cipher iv and key to " + configuration["conf_filename"]);
		}

		conf_file << std::endl;
		if (save_iv)
		{
			conf_file << "cipher_iv=" << configuration["cipher_iv"] << std::endl;
		}
		if (save_key)
		{
			conf_file << "cipher_key=" << configuration["cipher_key"] << std::endl;
		}
		conf_file << std::endl;
		conf_file.close();

		#ifdef EPS_WINDOWS
			/// @todo how to secure the configuration file in Windows?
		#else
			chmod(configuration["conf_filename"].c_str(), S_IRUSR	| S_IWUSR);
		#endif
	}

	return;
}


EPS::NetBuffer *EPS::Cipher::encrypt(EPS::NetBuffer &buffer)
{
	// figure out the PKCS #5 padding (add between 1 and 16 bytes to the end of the input)
	const size_t original_len				= buffer.length;
	const size_t number_of_padding_bytes	= EPS::number_of_pkcs_padding_bytes(original_len);
	const size_t full_len_with_padding		= original_len + number_of_padding_bytes;

	if (full_len_with_padding > EPS::network_buffer_size)
	{
		LOG_ERROR << full_len_with_padding << " bytes is too long to encrypt; output of first 64 bytes: " << EPS::hex_out(buffer.data(), 64);
		THROW_INVALID_ARGUMENT(-1, "attempt to encrypt too many bytes: " + std::to_string(original_len));
	}

	// set the PKCS padding bytes at the end of the buffer
	memset(buffer.data() + original_len, (int)number_of_padding_bytes, number_of_padding_bytes);

	EPS::NetBuffer *encrypted_buffer = buffer.create_related_buffer(EPS::NetBuffer::EType::kEncrypted);
	int tmplen = 0;

	// Encrypt the entire buffer at once.  Since the buffer is a multiple of an AES cipher block, no need to call EVP_CipherFinal_ex().
	EVP_CipherInit_ex(EPS::Cipher::encrypt_ctx, nullptr, nullptr, EPS::Cipher::key.data(), EPS::Cipher::iv.data(), ENC_MODE_UNCHANGED);
	EVP_CipherUpdate(EPS::Cipher::encrypt_ctx, encrypted_buffer->data(), &tmplen, buffer.data(), (int)full_len_with_padding);

	encrypted_buffer->length = tmplen;

	if (full_len_with_padding != (size_t)tmplen)
	{
		LOG_ERROR << "expected encrypted length to be " << full_len_with_padding << " bytes, but encrypted " << tmplen << " bytes instead";
		THROW_RUNTIME_ERROR(-1, "failed to finish encrypting " + std::to_string(original_len) + " bytes");
	}

	return encrypted_buffer;
}


EPS::NetBuffer *EPS::Cipher::decrypt_header(EPS::NetBuffer &buffer)
{
	const size_t original_length = buffer.length;

	// lie about the length so we only decrypt the first few bytes
	buffer.length = EPS::cipher_block_size;
	EPS::NetBuffer *output = EPS::Cipher::decrypt(buffer);

	buffer.length = original_length;

	return output;
}


EPS::NetBuffer *EPS::Cipher::decrypt(EPS::NetBuffer &buffer)
{
	const size_t len_encrypted = buffer.length;

	if (len_encrypted < EPS::cipher_block_size || (len_encrypted % EPS::cipher_block_size) != 0)
	{
		LOG_ERROR << "cannot decrypt due to invalid length: " << len_encrypted << " bytes: output of first 64 bytes: " << EPS::hex_out(buffer.data(), 64);
		THROW_INVALID_ARGUMENT(-1, "attempt to decrypt message of invalid length: " + std::to_string(len_encrypted));
	}

	if (len_encrypted > EPS::network_buffer_size)
	{
		LOG_ERROR << "cannot decrypt due to buffer overflow: output of first 64 bytes: " << EPS::hex_out(buffer.data(), 64);
		THROW_INVALID_ARGUMENT(-1, "attempt to decrypt too many bytes: " + std::to_string(len_encrypted));
	}

	EPS::NetBuffer::EType type = EPS::NetBuffer::EType::kFullyDecrypted;
	if (len_encrypted == EPS::cipher_block_size)
	{
		// looks like we're decrypting just the header
		type = EPS::NetBuffer::EType::kPartiallyDecrypted;
	}
	EPS::NetBuffer *decrypted_buffer = buffer.create_related_buffer(type);
	int tmplen = 0;

	// Decrypt the entire buffer at once.  Since the buffer is a multiple of an AES cipher block, no need to call EVP_CipherFinal_ex().
	EVP_CipherInit_ex(EPS::Cipher::decrypt_ctx, nullptr, nullptr, EPS::Cipher::key.data(), EPS::Cipher::iv.data(), ENC_MODE_UNCHANGED);
	EVP_CipherUpdate(EPS::Cipher::decrypt_ctx, decrypted_buffer->data(), &tmplen, buffer.data(), (int)len_encrypted);

	// If we're decrypting more than just the header, then we need to deal with the PKCS padding byte.
	if (type == EPS::NetBuffer::EType::kFullyDecrypted)
	{
		const uint8_t pad_byte = decrypted_buffer->data()[tmplen - 1];
		if (pad_byte < 1 || pad_byte > EPS::cipher_block_size)
		{
			LOG_ERROR << "truncating packet: pkcs padding byte value of " << (int)pad_byte << " (0x" << std::hex << (int)pad_byte << std::dec << ") is invalid while decrypting " << len_encrypted << " byte buffer";
			LOG_DEBUG << "output of first 64 decrypted bytes: " << EPS::hex_out(decrypted_buffer->data(), 64);
			decrypted_buffer->buffer_type = EPS::NetBuffer::EType::kInvalid;
			tmplen = 0;
		}
		else
		{
			tmplen -= (size_t)pad_byte;
		}
	}

	decrypted_buffer->length = tmplen;

	return decrypted_buffer;
}
