/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Hull.cpp 2194 2017-04-02 01:51:40Z stephane $
 * See license.txt.
 */

#include "Hull.hpp"
#include "Cipher.hpp"
#include "logging.hpp"
#include "common.hpp"
#include "comm.hpp"
#include <boost/program_options.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/uuid/nil_generator.hpp>
#include <boost/uuid/random_generator.hpp>
#include <boost/uuid/string_generator.hpp>
#include <boost/regex.hpp>
#include <chrono>
#include <exception>
#include <signal.h>
#include <typeinfo>
#include <fstream>
#include <iostream>
#include <sys/stat.h>
#include <sys/types.h>
#include <ifaddrs.h>


uint16_t	EPS::Hull::port				= 0;
std::string	EPS::Hull::addr				= "";
std::string	EPS::Hull::network_str		= "";
std::string	EPS::Hull::scorecard		= "";
std::string	EPS::Hull::conf_filename	= "/etc/eps.conf";
std::string EPS::Hull::log_filename		= "";
std::string EPS::Hull::gv_filename		= "";
bool		EPS::Hull::sequential		= false;
bool		EPS::Hull::upnp_enabled		= false;
int			EPS::Hull::upnp_timeout		= 1000;		///< in milliseconds


static EPS::Hull *ptr = nullptr;

EPS::Hull &EPS::Hull::get(void)
{
	if (ptr == nullptr)
	{
		THROW_RUNTIME_ERROR(-1, "Cannot de-reference null EPS hull pointer." );
	}

	return *ptr;
}


EPS::Hull::Hull(const EPS::EMode m, int argc, char **argv) :
	mode					(m						),
	mode_string				(EPS::mode_to_string(m)	),
	done					(false					),
	ac						(argc					),
	av						(argv					),
	seconds_between_retries	(EPS::get_random(20, 60)),
	uuid					(boost::uuids::nil_uuid()),
	io_service				(1						),
	signal_set				(io_service				),
	tcp_resolver			(io_service				),
	acceptor				(io_service				),
	connect_timer			(io_service				),
	stats_timer				(io_service				),
#ifndef EPS_SKIP_UPNP
	upnp_renew_timer		(io_service				),
	upnp_wan_port			(0						),
	upnp_dev				(nullptr				),
#endif
	start_time				(std::time(nullptr)		)
{
	ptr = this;

	EPS::initialize_logging(mode);
	LOG_INFO << "starting " << EPS_SYSTEM << " EPS " << mode_string << " v" << EPS_VERSION;

#ifndef EPS_SKIP_UPNP
	memset(&upnp_urls, '\0', sizeof(upnp_urls));
	memset(&upnp_data, '\0', sizeof(upnp_data));
#endif

	LOG_DEBUG << "inner " << mode_string << " hull constructed";

	return;
}


EPS::Hull::~Hull(void)
{
	LOG_DEBUG << "inner " << mode_string << " hull destruction in progress";

	this->done = true;
	this->stop_asio();

	EPS::Cipher::reset();

	if (std::time(nullptr) - this->start_time > 1)
	{
		LOG_INFO << "EPS " << mode_string << " ran for " << EPS::get_approximate_time(this->start_time);
	}
	LOG_INFO << "stopping " << EPS_SYSTEM << " EPS " << mode_string << " v" << EPS_VERSION;

	ptr = nullptr;

	return;
}


void EPS::Hull::stop_asio(void)
{
	this->done = true;

#ifndef EPS_SKIP_UPNP
	upnp_renew_timer.cancel(this->ec);
#endif
	connect_timer	.cancel(this->ec);
	stats_timer		.cancel(this->ec);
	acceptor		.cancel(this->ec);
	tcp_resolver	.cancel(		);

	// close all connections
	for (EPS::TConnPtr connection : this->tracker.get_all_connections())
	{
		this->tracker.deactivate_entry(connection->uuid);
	}

	return;
}


void EPS::Hull::start_signal_handling(void)
{
	// called by EPS::Hull::run_first_stage()

	LOG_TRACE << "setting up signal handling";

	signal_set.add( SIGABRT	);	// abort
	signal_set.add( SIGINT	);	// "CTRL+C"
#ifdef EPS_WINDOWS
	signal_set.add( SIGBREAK);	// "CTRL+BREAK"
#else
	signal_set.add( SIGQUIT	);	// "CTRL+\"
#endif
	signal_set.add( SIGILL	);	// illegal instruction
	signal_set.add( SIGSEGV	);	// segfault
	signal_set.add( SIGTERM	);	// terminate

	signal_set.async_wait(
			[this](const EC &error, int signal_number)
			{
				// Lambda: signal handler

				if (not error)
				{
					std::string signal_name = "signal #" + std::to_string(signal_number);
					#ifndef EPS_WINDOWS
						signal_name += " (" + std::string(strsignal(signal_number)) + ")";
					#endif
					LOG_FATAL << "FATAL!  Signal handler called for " << signal_name << ".";
					this->ec = boost::system::error_code(-1, boost::system::generic_category()); /// @todo define our own error category derived from error_category
					this->done = true;
					this->io_service.stop();
					LOG_WARN << "io service has been stopped due to " << signal_name;

					/* These are all fatal signals, so don't bother re-installing a new signal handler.
					 *
					start_signal_handling();
					 *
					 */
				}

				return;
			} );

	return;
}


void EPS::Hull::setup_encryption_and_decryption(void)
{
	if (this->mode == EPS::kClient)
	{
		/* Servers create their own key and iv if they're missing from the .conf file, but clients cannot do that.
		 * The clients must be explicitly told which key and iv to use, and it must match the server to which they're
		 * connecting.  If they key and iv doesn't exist or isn't the right size, then immediately abort since nothing
		 * can possibly work until the problem is resolved.
		 */
		bool fail = false;

		if (this->configuration["cipher_iv"].size() != 2 * EPS::Cipher::iv.max_size())
		{
			LOG_ERROR << "cipher_iv must be " << EPS::Cipher::iv.max_size() << " bytes in length or " << EPS::Cipher::iv.max_size()*2 << " hex digits (actual: " << this->configuration["cipher_iv"].size() << " characters)";
			fail = true;
		}
		if (this->configuration["cipher_key"].size() != 2 * EPS::Cipher::key.max_size())
		{
			LOG_ERROR << "cipher_key must be " << EPS::Cipher::key.max_size() << " bytes in length or " << EPS::Cipher::key.max_size()*2 << " hex digits (actual: " << this->configuration["cipher_key"].size() << " characters)";
			fail = true;
		}
		if (fail)
		{
			LOG_ERROR << "Please copy cipher_iv and cipher_key values from the server's eps.conf configuration file to enable encryption/decryption.";
			THROW_INVALID_ARGUMENT(-1, "failed to import required encryption settings from " + this->conf_filename);
		}
	}

	EPS::Cipher::initialize(this->configuration);

	return;
}


void EPS::Hull::get_uuid(void)
{
	const std::string key = this->mode_string + "_uuid";

	if (this->configuration.count(key))
	{
		auto gen = boost::uuids::string_generator();
		this->uuid = gen(this->configuration[key]);
		LOG_INFO << "imported previous " << this->mode_string << " UUID: " << uuid;
	}

	if (uuid.is_nil())
	{
		auto gen = boost::uuids::random_generator();
		uuid = gen();
		LOG_INFO << "generated a new " << this->mode_string << " UUID: " << uuid;

		this->save_uuid();
	}

	return;
}


void EPS::Hull::save_uuid(void)
{
	std::ofstream conf_file(this->conf_filename, std::ofstream::out | std::ofstream::app);
	if (not conf_file)
	{
		THROW_RUNTIME_ERROR(errno, "Cannot save UUID to " + conf_filename);
	}

	conf_file	<< std::endl
				<< this->mode_string << "_uuid=" << uuid << std::endl
				<< std::endl;
	conf_file.close();

	#ifdef EPS_WINDOWS
		/// @todo how to fix this on Windows?
	#else
		chmod(this->conf_filename.c_str(), S_IRUSR	| S_IWUSR);
	#endif

	return;
}


void EPS::Hull::write_blank_configuration_file(void)
{
	std::ofstream of(this->conf_filename);
	if (not of)
	{
		LOG_ERROR << "cannot write to configuration file " << conf_filename;
		return;
	}

	of	<< "# EPS configuration file originally created by "
		<< "EPS v"	<< EPS_VERSION
		<< " on "	<< EPS::get_utc_timestamp() << "."  << std::endl
		<< ""											<< std::endl
		<< ""											<< std::endl
		<< EPS::get_eps_conf_template()					<< std::endl;

	of.close();

	#ifdef EPS_WINDOWS
		/// @todo how to fix this on Windows?
	#else
		chmod(this->conf_filename.c_str(), S_IRUSR | S_IWUSR);
	#endif

	LOG_WARN	<< ""																				<< std::endl
				<< "-----------------------------"													<< std::endl
				<< "A new configuration file has been saved as "	<< this->conf_filename	<< "."	<< std::endl
				<< "Please edit this file and restart the EPS "		<< this->mode_string	<< "."	<< std::endl
				<< "-----------------------------"													<< std::endl;

	return;
}


void EPS::Hull::read_configuration_file(void)
{
	this->configuration.clear();
	this->configuration["conf_filename"] = this->conf_filename;

	std::ifstream conf_file(this->conf_filename);
	if (not conf_file)
	{
		LOG_WARN << "cannot read from configuration file " << conf_filename;
		this->write_blank_configuration_file();
	}
	else
	{
		/* Read all entries from the configuration file.  Entries we'll keep look like this:
		 *
		 *		key=value
		 *		key = value
		 *
		 * Everything else is ignored.
		 */

		const boost::regex rx(
			"^"									// start of line
			"\\s*"								// remove leading whitespace
			"([-_A-Za-z0-9]+)"					// key
			"\\s*=\\s*"							// =
			"([^#]+?)"							// value (non-greedy so the next whitespace can match)
			"\\s*"								// trailing whitespace
			"(#.*)?"							// trailing comment (ignored)
			"$"									// end of line
			);

		std::string line;
		while (std::getline(conf_file, line))
		{
			boost::smatch what;
			if (not boost::regex_match(line, what, rx))
			{
				// line doesn't match
				continue;
			}

			const std::string key = what[1].str();
			const std::string val = what[2].str();

			if (this->configuration.count(key))
			{
				LOG_WARN	<< "found duplicate configuration entry:"				<< std::endl
							<< "\t"		<< key << "=" << this->configuration[key]	<< std::endl
							<< "or:\t"	<< key << "=" << val;
			}
//			LOG_TRACE << "imported " << key << "=" << val; -- security risk! (exposes cipher iv and key in log files)
			this->configuration[key] = val;
		}

		LOG_INFO << "imported " << this->configuration.size() << " configuration items from " << this->conf_filename;
	}

	return;
}


void EPS::Hull::parse_options(void)
{
	/** For historical reasons, not all of the configuration settings are read/parsed using boost::program_options.
	 * For example, see the UUID values as well as cipher_iv and cipher_key in EPS::Hull::read_configuration_file().
	 */

	boost::program_options::options_description	common_options	("Common options"			);
	boost::program_options::options_description	log_stat_options("Logging and statistics"	);
	boost::program_options::options_description	client_options	("Client options"			);
	boost::program_options::options_description	server_options	("Server options"			);
	boost::program_options::options_description	all				("EPS options"				);
	boost::program_options::variables_map		vm;

	common_options.add_options()
			("conf"				, boost::program_options::value<std::string>(&this->conf_filename)->default_value("/etc/eps.conf"), "filename from which additional EPS configuration will be read")
			("help"				, "show this help message"		)
			("version"			, "show version information"	)
			;
	all.add(common_options);

	log_stat_options.add_options()
			("log"				, boost::program_options::value<std::string>(&this->log_filename)->default_value("/var/log/eps_" + this->mode_string + ".log"), "file to which logging messages will be written (file is auto-rotated)")
			("stats"			, "log statistics every 60 seconds")
			("scorecard"		, boost::program_options::value<std::string>(&this->scorecard)->implicit_value("/var/log/eps_" + this->mode_string + ".stats"), "file where statistics scorecard is written every 60 seconds")
			;
	all.add(log_stat_options);

	if (this->mode == EPS::kServer)
	{
		server_options.add_options()
				("graphviz"		, boost::program_options::value<std::string>(&this->gv_filename	)->implicit_value("/var/log/eps.gv"	), "file to which a Graphviz-format network diagram is written every 5 minutes")
				("network"		, boost::program_options::value<std::string>(&this->network_str	)->default_value("192.168.43.0/24"	), "network (with mask) to create"	)
				("port"			, boost::program_options::value<uint16_t>	(&this->port		)->default_value(43439				), "TCP/UDP listening port to bind"	)
				("sequential"	, "new clients will be assign IP addresses sequentially (.1, .2, .3, etc) instead of from within the pool")
				("upnptimeout"	, boost::program_options::value<int>(&this->upnp_timeout)->default_value(1000), "UPnP timeout in milliseconds (set to zero to disable UPnP support)")
				;
		all.add(server_options);
	}

	if (this->mode == EPS::kClient)
	{
		client_options.add_options()
				("port"			, boost::program_options::value<uint16_t>	(&this->port)->default_value(43439), "TCP/UDP port of the server")
				("server"		, boost::program_options::value<std::string>(&this->addr), "IP address (or hostname) of the server")
//				("upnptimeout"	, boost::program_options::value<int>(&this->upnp_timeout)->default_value(1000), "UPnP timeout in milliseconds (set to zero to disable UPnP support)")
				;
		all.add(client_options);
	}

	boost::program_options::store(boost::program_options::parse_command_line(ac, av, all), vm);
	boost::program_options::notify(vm);

	std::ifstream conf_file(this->conf_filename);
	if (conf_file)
	{
		LOG_INFO << "reading EPS configuration file " << this->conf_filename;
		boost::program_options::store(boost::program_options::parse_config_file(conf_file, all, true), vm);
		boost::program_options::notify(vm);
	}
	else
	{
		LOG_WARN << "failed to read EPS configuration file " << this->conf_filename;
	}

	this->log_filename = vm["log"].as<std::string>();
	EPS::initialize_file_logging(this->log_filename);

	if (vm.count("version"))
	{
		this->done = true;
		std::cout	<< EPS_SYSTEM << " EPS " << this->mode_string << " v" << EPS_VERSION	<< std::endl
					<< "Copyright 2014-2017 Stephane Charette <stephanecharette@gmail.com>"		<< std::endl;
	}
	else if (vm.count("help"))
	{
		this->done = true;
		std::cout << all << std::endl;
	}
	else
	{
		if (this->mode == EPS::kClient)
		{
			if (vm.count("server") == 0)
			{
				THROW_INVALID_ARGUMENT(-1, "Client requires the \"--server\" option to specify the server's name or ip address.");
			}
			LOG_INFO << "server address is set to " << vm["server"].as<std::string>();
		}
		else if (this->mode == EPS::kServer)
		{
			EPS::parse_network_and_mask(vm["network"].as<std::string>(), this->network, this->netmask);

			if (vm.count("sequential"))
			{
				LOG_WARN << "client ip addresses will be assigned sequentially";
				this->sequential = true;
			}

			if (vm.count("graphviz"))
			{
				LOG_INFO << "graphviz network diagram is enabled: " << this->gv_filename;
			}
			else
			{
				LOG_INFO << "graphviz network diagram is disabled (use \"--graphviz\" to enable)";
			}
		}

		if (vm.count("scorecard"))
		{
			this->stats.scorecard_enabled	= true;
			this->stats.scorecard_filename	= vm["scorecard"].as<std::string>();
			LOG_INFO << "statistics scorecard is enabled: " << this->stats.scorecard_filename;
		}
		else
		{
			this->stats.scorecard_enabled	= false;
			LOG_INFO << "statistics scorecard is disabled (use \"--scorecard\" to enable)";
		}

		if (vm.count("stats"))
		{
			this->stats.log_stats = true;
			LOG_INFO << "statistics logging is enabled";
		}
		else
		{
			this->stats.log_stats = false;
			LOG_INFO << "statistics logging is disabled (use \"--stats\" to enable)";
		}

		if (vm.count("upnptimeout"))
		{
			this->upnp_timeout = vm["upnptimeout"].as<int>();
			if (this->upnp_timeout > 0)
			{
				this->upnp_enabled = true;
				LOG_INFO << "UPnP timeout: " << this->upnp_timeout << " milliseconds";
			}
			else
			{
				this->upnp_enabled = false;
				LOG_INFO << "UPnP support is disabled";
			}
		}
	}

	return;
}


void EPS::Hull::run_first_stage(void)
{
	// setup the top-level try/catch block so we don't rewind past main()
	try
	{
		this->start_signal_handling();

		#ifdef EPS_WINDOWS
			/// @todo how to check for administrator?
		#else
			if (getuid()	!= 0 &&
				getgid()	!= 0 &&
				geteuid()	!= 0 &&
				getegid()	!= 0 )
			{
				LOG_WARN << "EPS is not running as root -- this will likely cause permission problems and prevent EPS from running correctly";
			}
		#endif

		if (not done) this->parse_options();
		if (not done) this->read_configuration_file();
		if (not done) this->get_uuid();
		if (not done) this->setup_encryption_and_decryption();
		if (not done) this->start_stats_timer();
		if (not done) this->stats.roll();
		if (not done) this->run_second_stage();

		// attempt to do a clean shutdown
		this->io_service.reset();
		this->stop_asio();
		this->io_service.poll(this->ec);
	}
	catch (const std::exception &e)
	{
		LOG_FATAL << "FATAL!  Exiting due to " << EPS::cpp_demangle(typeid(e).name()) << " exception:  " << e.what();
	}
	catch (...)
	{
		LOG_FATAL << "FATAL!  Exiting due to unknown exception.";
	}

	io_service.stop();

	return;
}


void EPS::Hull::connection_failed( EPS::TCPConnection *connection )
{
	// Note this is being called with a raw TCPConnection pointer object, not a shared pointer.
	// This likely means the problem happened very early in the connection process, or within
	// the TCPConnection object itself.

	EPS::TConnPtr ptr;
	try
	{
		ptr = this->tracker.get_shared_pointer(connection);
	}
	catch(...)
	{
		// nothing we can do about it, looks like we don't have a shared pointer for this connection
		LOG_WARN << connection->name << "ignoring connection since it wasn't yet tracked";
	}

	this->connection_failed(ptr);

	return;
}


void EPS::Hull::readComm( const EPS::TConnPtr &connection )
{
	// Beware that if the connection is still being validated (meaning "Hello" hasn't yet been received)
	// then this connection does **NOT** yet exist in the tracker.

	if (this->done)
	{
		return;
	}

//	LOG_TRACE << connection->name << "waiting to read the next comm";

	EPS::NetBuffer *nb = new EPS::NetBuffer(EPS::NetBuffer::EType::kEncrypted);

	boost::asio::async_read( connection->socket, nb->buffer_1st, boost::asio::transfer_exactly(EPS::cipher_block_size),
			[=](const EC &error, size_t bytes_read_1)
			{
				if (error || bytes_read_1 != EPS::cipher_block_size)
				{
					LOG_ERROR << connection->name << "failed to read communication command: " << error.message();
					this->connection_failed(connection);
					delete nb;
					return;
				}

				// we've read exactly 1 block, so decrypt the packet header
				nb->length = bytes_read_1;
				const EPS::CommHeader1 *header = nb->get_header();
				const size_t total_number_of_bytes = header->size_of_packet();

				if (header->protocol_version != EPS_PROTOCOL_VERSION)
				{
					LOG_ERROR << connection->name << "peer is using EPS protocol v" << (int)header->protocol_version << " which is not compatible with our own EPS protocol v" << (int)EPS_PROTOCOL_VERSION;
					this->connection_failed(connection);
					delete nb;
					return;
				}

				if (header->comm_type <= EPS::ECommType::kUnknown	||
					header->comm_type >= EPS::ECommType::kMax		)
				{
					// this cannot be a valid packet since the comm type is unrecognized
					LOG_ERROR	<< connection->name << "invalid communication type: #" << (int)header->comm_type;
					LOG_WARN	<< connection->name << "possible cause: has the cipher key and iv been setup correctly in " << this->conf_filename << "?";
					this->connection_failed(connection);
					delete nb;
					return;
				}

				if (total_number_of_bytes <= sizeof(EPS::CommHeader1)	||
					total_number_of_bytes > EPS::network_buffer_size	)
				{
					LOG_ERROR << connection->name << "invalid number of bytes in communication command: " << total_number_of_bytes;
					this->connection_failed(connection);
					delete nb;
					return;
				}

				const size_t number_of_bytes_remaining_to_read = EPS::length_including_pkcs_padding(total_number_of_bytes) - bytes_read_1;

				boost::asio::async_read( connection->socket, nb->buffer_2nd, boost::asio::transfer_exactly(number_of_bytes_remaining_to_read),
						[=](const EC &err, size_t bytes_read_2)
						{
							const size_t total_bytes_read = bytes_read_1 + bytes_read_2;
							nb->length = total_bytes_read;

							if (connection->is_mesh)
							{
								this->stats.minute_stats.direct_bytes_received		+= total_bytes_read;
								this->stats.minute_stats.direct_packets_received	++;
							}
							else
							{
								this->stats.minute_stats.server_bytes_received		+= total_bytes_read;
								this->stats.minute_stats.server_packets_received	++;
							}

							if (err || bytes_read_2 != number_of_bytes_remaining_to_read || this->done)
							{
								LOG_ERROR << connection->name << "failed to read communication command (read " << bytes_read_2 << " of " << number_of_bytes_remaining_to_read << " bytes): " << err.message();
								this->connection_failed(connection);
								delete nb;
								return;
							}

							// otherwise, if we get here then we should have a full command to interpret
							this->processComm( connection, nb->get_decrypted() );
						} );
			} );

	return;
}


void EPS::Hull::processComm( const EPS::TConnPtr &connection, EPS::NetBuffer *nb )
{
	// Beware that if the connection is still being validated (meaning "Hello" hasn't yet been received)
	// then this connection does **NOT** yet exist in the tracker.

	if (nb == nullptr)
	{
		// What in the world happened?  This shouldn't ever happen.
		LOG_ERROR << connection->name << "cannot process null NetBuffer";
		this->connection_failed(connection);
		return;
	}

	const EPS::CommHeader1	*header		= nb	->get_header();
	const EPS::ECommType	type		= header->type();

//	LOG_TRACE << connection->name << "processing " << header->size_of_packet() << " bytes, comm type #" << (int)type << " (" << EPS::comm_type_to_string(type) << ")";

	if (type == EPS::kHello2)
	{
		/* This is the first message sent between eps endpoints.  The server side will eventually reply with
		 * EPS::kAddress1.  But until we reach the end of this if() block, the tracker doesn't know anything
		 * about this connection.
		 */

		const EPS::CommHello2 *ch2 = reinterpret_cast<const EPS::CommHello2*>(header);
		if (EPS::Cipher::hash != ch2->cipherhash_str)
		{
			LOG_ERROR	<< connection->name << "local encryption/decryption cipher key and iv hash " << EPS::Cipher::hash << " does not match peer hash " <<  ch2->cipherhash_str;
			LOG_WARN	<< connection->name << "connection will be closed due to previous error";
			LOG_INFO	<< connection->name << "verify the cipher settings in " << this->conf_filename << " and try again";
			this->connection_failed(connection);
			return;
		}

		const std::string our_eps_version(EPS_VERSION);
		if (our_eps_version != ch2->version_str)
		{
			// if we got this far, we know the protocol version numbers are valid, but at least log a warning about the different EPS version numbers
			LOG_WARN << connection->name << "peer version string " << ch2->version_str << " does not match local version string " << our_eps_version;
			LOG_WARN << connection->name << "our local version of EPS is " << (our_eps_version < ch2->version_str ? "older" : "newer") << " than the peer; please consider upgrading";
		}

		auto uuid_generator = boost::uuids::string_generator();
		const UUID uuid = uuid_generator(ch2->uuid_str);
		if (not connection->uuid.is_nil() && connection->uuid != uuid)
		{
			LOG_ERROR	<< connection->name << "uuid already assigned to this connection (" << connection->uuid << ") but new hello2 received from " << EPS::mode_to_string(static_cast<EPS::EMode>(ch2->mode)) << " with uuid " << uuid;
			LOG_WARN	<< connection->name << "connection will be closed due to previous error";
			this->connection_failed(connection);
			return;
		}

		const bool already_exists = this->tracker.exists(uuid);

		// if both us and the connecting peer are eps clients...
		if (ch2->mode == EPS::kClient && this->mode == EPS::kClient)
		{
			if (this->server_conns.empty())
			{
				// forget about trying to connect directly to a client if we don't yet have a server connection!
				LOG_ERROR	<< connection->name << "'hello2' comes from a client -- need to connect to a server first";
				LOG_WARN	<< connection->name << "connection will be closed due to previous error";
				this->connection_failed(connection);
				return;
			}

			if (not already_exists)
			{
				// we shouldn't accept a wild peer connection from a device that our server hasn't told us about yet
				LOG_ERROR	<< connection->name << "refusing connection from unknown device (peer is not in our known list of peers)";
				LOG_WARN	<< connection->name << "connection will be closed due to previous error";
				this->connection_failed(connection);
				return;
			}

			if (this->tracker.exists(uuid))
			{
				EPS::TrackerData &data = this->tracker.get(uuid);

				if (data.pending_mesh_connection && data.pending_mesh_connection != connection)
				{
					// close the old pending connection now that we seem to have a new connection
					LOG_WARN << connection->name << "is replacing the previous pending connections to " << data.as_debug();
					this->connection_failed(data.pending_mesh_connection);
				}

				if (data.connection && data.connection != connection)
				{
					// close the old direct connection now that we seem to have an new connection
					LOG_WARN << connection->name << "is replacing the previous direct connections to " << data.as_debug();
					this->connection_failed(data.connection);
				}
			}
		}

		// if we get here, then the 'hello2' request looks good, and we'll insert (or update) the tracker record

		const IPv4 ipv4_eps(ntohl(ch2->address_eps));	// this one may be a null address if the client is connecting to the server
		EPS::TrackerData &data = this->tracker.add_active_entry(uuid, ipv4_eps, connection);
		data.hostname		= ch2->hostname_str;
		data.os_name		= ch2->os_version_str;
		data.eps_version	= ch2->version_str;
		data.eps_uptime		= ntohl(ch2->eps_uptime);
		data.system_uptime	= ntohl(ch2->system_uptime);
		data.mode 			= static_cast<EPS::EMode>(ch2->mode);
		data.msg			= "valid " + EPS::mode_to_string(data.mode) + " connection";

		for (size_t idx = 0; idx < MAX_NUMBER_OF_OTHER_ADDRESSES; idx ++)
		{
			IPv4 ip(ntohl(ch2->other_addresses[idx]));
			if (ip.is_unspecified() == false)
			{
				data.other_addresses.insert( ip );
			}
		}
		data.other_addresses.insert( connection->socket.remote_endpoint().address().to_v4() );

		LOG_INFO << connection->name << "received 'hello2' from uuid " << uuid << " and the following addresses: " << ipset_to_string(data.other_addresses);
	}

	// see if the client/server has some special handling it needs to do with this communication
	processComm(connection, nb, header, type);

	delete nb;
	nb = nullptr;

	// get ready to read the next command
	readComm(connection);

	return;
}


void EPS::Hull::validate_new_connection( const EPS::TConnPtr &connection )
{
	// Note that when this is called, the connection has **NOT YET** been inserted into the tracker.
	// That isn't done until the "Hello" message has been received and we know the UUID for certain.

	LOG_DEBUG << connection->name << "validating new TCP beam";

	this->stats.minute_stats.connections_established ++;

	EPS::NetBuffer	*nb		= new EPS::NetBuffer(EPS::NetBuffer::EType::kFullyDecrypted);
	EPS::CommHello2 *ch2	= reinterpret_cast<EPS::CommHello2*>(nb->data());

	// enumerate and send all of our interesting local addresses
	IPv4Set s;
	ifaddrs *addrs = nullptr;
	getifaddrs(&addrs);
	for (ifaddrs *ifa = addrs; ifa != nullptr; ifa = ifa->ifa_next)
	{
		if (ifa->ifa_addr == nullptr)
		{
			// skip interfaces with no information
			continue;
		}

		if (ifa->ifa_addr->sa_family != AF_INET)
		{
			// skip non-ipv4 family
			continue;
		}

		if (boost::regex_match(ifa->ifa_name, boost::regex("^eps[0-9]+$")))
		{
			// skip other EPS addresses
			continue;
		}

		const IPv4 ip( htonl(reinterpret_cast<sockaddr_in *>(ifa->ifa_addr)->sin_addr.s_addr) );
		if (ip.is_loopback() || ip.is_multicast() || ip.is_unspecified())
		{
			// skip useless (to us) addresses
			continue;
		}

		// finally found an interesting address
		s.insert( ip );
	}

	if (addrs)
	{
		freeifaddrs(addrs);
	}

	// whatever local address we're using should already be in the set, but just in case something went wrong
	s.insert( connection->socket.local_endpoint().address().to_v4() );

	local_addresses = s;

	ch2->set(this->uuid, this->mode, this->start_time, EPS::Cipher::hash, s, this->address);
	nb->length = ch2->header.size_of_packet();

	LOG_TRACE << connection->name << "sending initial 'hello2' with the following addresses: " << ipset_to_string(s);

	this->async_write(connection, nb);
	this->readComm(connection);

	return;
}


void EPS::Hull::validate_new_connection( EPS::TCPConnection *connection )
{
	/* A raw TCP connection object has been created.  This can either be an existing server connection which is
	 * in the process of being established, or we have a brand connection to a client.  Get (or create) a shared
	 * pointer to this connection and then kick off the process of validating the new connection.
	 */
	EPS::TConnPtr ptr;

	// is this a connection to a server?
	for (const EPS::TConnPtr &conn : this->server_conns)
	{
		if (conn.get() == connection)
		{
			ptr = conn;
			break;
		}
	}

	// ...if not, then it must be a connection to a client
	if (ptr.get() == nullptr)
	{
		ptr = this->tracker.get_shared_pointer(connection);

		if (this->mode == EPS::kClient)
		{
			// this is a client-to-client connection (mesh)
			ptr->is_mesh = true;
		}
	}

	this->validate_new_connection( ptr );

	return;
}


void EPS::Hull::start_connect_timer(void)
{
	const int seconds = EPS::get_random(30, 300);	// wait a random amount of time so all clients don't trigger at the exact same time
//	const int seconds = EPS::get_random(10, 25);

	LOG_TRACE << "next mesh connection timer is in " << seconds << " seconds";

	this->connect_timer.expires_from_now(std::chrono::seconds(seconds));
	this->connect_timer.async_wait(
			[=](const EC &error)
			{
				if (not error && not this->done)
				{
					// go through the connections and see if we can establish anything new
					const std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
					const auto server_address = this->server_conns.begin()->get()->endpoint.address();

					LOG_DEBUG << "number of UUIDs to which we'd like to establish a new mesh connection: " << this->pending_uuids.size();

					for (const UUID &uuid : this->pending_uuids)
					{
						if (this->tracker.does_not_exist(uuid))
						{
							LOG_WARN << "cannot consider a mesh connection since UUID seems to no longer exist: " << uuid;
						}
						else
						{
							EPS::TrackerData &data = this->tracker.get(uuid);

							// see if we should try to connect to this peer
							data.msg.clear();
								 if (this->uuid == uuid)						data.msg = "mesh: not attempting to connect to self";
							else if (data.mode == EPS::kServer)					data.msg = "mesh: cannot connect to a server";
							else if (data.mode != EPS::kClient)					data.msg = "mesh: cannot connect to a non-client endpoint";
							else if (data.connection != nullptr)				data.msg = "mesh: direct connection already exists";
							else if (data.pending_mesh_connection != nullptr)	data.msg = "mesh: pending connection already exists";
							else if (data.next_connection_attempt > now)		data.msg = "mesh: must wait longer before re-trying to connect";

							if (not data.msg.empty())
							{
								// we found a reason why we shouldn't connect
								continue;
							}

							if (data.mesh_addresses.empty())
							{
								LOG_INFO << "set of mesh address is empty, so importing all known \"other addresses\" for this entry: " << ipset_to_string(data.other_addresses);
								data.mesh_addresses = data.other_addresses;
							}

							/* If a number of devices are behind a NAT, then they'll all appear to have the exact
							 * same IP address to the server.  Since we're trying to connect to someone else, start
							 * by removing all IP addresses we have in common.
							 */
							for (const auto local_ip : this->local_addresses)
							{
								data.mesh_addresses.erase(local_ip);
							}

							// find an address to use to create a direct mesh connection
							EPS::IPv4 ip;
							if (not data.mesh_addresses.empty())
							{
								// grab the next address to which we want to connect
								ip = *data.mesh_addresses.begin();
								data.mesh_addresses.erase(ip);
								LOG_INFO << "mesh connection timer: IP address to which we should try to connect: " << ip.to_string() << ", remaining addresses are: " << ipset_to_string(data.mesh_addresses);
							}

							// do a few last-minute checks on the address we're about to use
								 if (ip.is_unspecified())						data.msg = "mesh: cannot connect to empty address";
							else if (ip.is_loopback())							data.msg = "mesh: cannot connect to loopback address";
							else if (ip == server_address)						data.msg = "mesh: cannot connect to the server's address";
							else
							{
								data.msg = "mesh: creating new mesh connection to " + ip.to_string();

								const EPS::TCPEP tcp_endpoint(ip, this->port);
								EPS::TConnPtr connection(new EPS::TCPConnection(tcp_endpoint, uuid));
								LOG_INFO << connection->name << "attempting to connect to " << uuid << " at " << tcp_endpoint;
								EPS::TrackerData &data = this->tracker.get(uuid);
								data.pending_mesh_connection = connection;
								connection->async_connect();

								// set a long "retry" time before we'll make another attempt
								data.next_connection_attempt = now + std::chrono::minutes(EPS::get_random(5, 45));
//								data.next_connection_attempt = now + std::chrono::minutes(1);
							}

							LOG_INFO << "mesh connection timer result: " << data.as_debug();
						}
					}

					this->pending_uuids.clear();
					// schedule the next timer
					this->start_connect_timer();
				}
			} );

	return;
}


void EPS::Hull::send_stats(void)
{
	if (not this->done && this->server_conns.empty() == false)
	{
		EPS::TConnPtr conn = *this->server_conns.begin();

		EPS::NetBuffer	*nb		= new EPS::NetBuffer(EPS::NetBuffer::EType::kFullyDecrypted);
		EPS::CommStats1	*cs1	= reinterpret_cast<EPS::CommStats1*>(nb->data());

		EPS::UUIDSet uuids = this->tracker.get_all_active_uuids();
		cs1->set(uuids);
		nb->length = cs1->header.size_of_packet();

		cs1->server_bytes_received	= this->stats.total_stats.server_bytes_received;
		cs1->server_bytes_sent		= this->stats.total_stats.server_bytes_sent;
		cs1->bytes_dropped			= this->stats.total_stats.bytes_dropped;
		cs1->direct_bytes_received	= this->stats.total_stats.direct_bytes_received;
		cs1->direct_bytes_sent		= this->stats.total_stats.direct_bytes_sent;
		cs1->tap_bytes_received		= this->stats.total_stats.tap_bytes_received;
		cs1->tap_bytes_sent			= this->stats.total_stats.tap_bytes_sent;

		strncpy(cs1->server_bps_in	, this->stats.total_stats.server_bps_in	, EPS::bps_string_size);
		strncpy(cs1->server_bps_out	, this->stats.total_stats.server_bps_out, EPS::bps_string_size);
		strncpy(cs1->direct_bps_in	, this->stats.total_stats.direct_bps_in	, EPS::bps_string_size);
		strncpy(cs1->direct_bps_out	, this->stats.total_stats.direct_bps_out, EPS::bps_string_size);
		strncpy(cs1->tap_bps_in		, this->stats.total_stats.tap_bps_in	, EPS::bps_string_size);
		strncpy(cs1->tap_bps_out	, this->stats.total_stats.tap_bps_out	, EPS::bps_string_size);

		LOG_TRACE << conn->name << "sending recent statistics up to the server";
		this->async_write(conn, nb);
	}

	return;
}


void EPS::Hull::start_stats_timer(void)
{
	this->stats_timer.expires_from_now(std::chrono::seconds(60));
	this->stats_timer.async_wait(
			[=](const EC &error)
			{
				if (not error && not this->done)
				{
					// combine the last 60-second stats into the totals, and remember the most recent BPS value
					this->stats.roll();

					if (this->mode == EPS::kClient)
					{
						// every five minutes, send our stats up to the server
						#ifdef NDEBUG
							const int send_stats_in_minutes = 5;
						#else
							const int send_stats_in_minutes = 1;
						#endif

						static int counter = 0;
						counter ++;
						if (counter >= send_stats_in_minutes)
						{
							this->send_stats();
							counter = 0;
						}
					}

					// schedule the next stats timer
					this->start_stats_timer();
				}
			} );

	return;
}


void EPS::Hull::setup_listener(void)
{
	const TCPEP endpoint( boost::asio::ip::tcp::v4(), this->port );

	this->acceptor.close(ec);
	this->acceptor.open(endpoint.protocol(), ec);
	if (ec)
	{
		THROW_RUNTIME_ERROR(errno, "Failed to open listener socket: " + ec.message());
	}

	this->acceptor.set_option( boost::asio::socket_base::reuse_address(true), ec );
	this->acceptor.bind(endpoint, ec);
	if (ec)
	{
		LOG_ERROR << "failed to bind to endpoint TCP:" << endpoint << ": " << ec.message();
		if (this->mode == EPS::kClient)
		{
			LOG_WARN << "this client can only make outgoing mesh connections due to the previous error";
			ec.clear();
		}
	}
	else
	{
		this->acceptor.listen(5, ec);
		if (ec)
		{
			LOG_ERROR << "failed to listen on " << endpoint << ": " << ec.message();
		}
		else
		{
			this->start_async_accept();
			LOG_INFO << "tcp listener waiting for inbound beams on " << endpoint;
		}
	}

	return;
}


void EPS::Hull::start_async_accept(void)
{
	TConnPtr accepted_connection(new EPS::TCPConnection);
	LOG_TRACE << accepted_connection->name << "waiting for incoming TCP connection";

	this->acceptor.async_accept(accepted_connection->socket,
			[=](const EC &error)
			{
				if (error || this->done)
				{
					LOG_WARN << accepted_connection->name << "failed to accept tcp connection: " << error.message();
					stats.minute_stats.connections_dropped ++;
				}
				else if (this->mode == EPS::kServer && this->ipv4_pool.empty())
				{
					/// @todo fixme: even if the pool is empty, if this client has a reserved ip address we still need to accept it
					LOG_WARN << accepted_connection->name << "cannot accept new connection from " << accepted_connection->socket.remote_endpoint() << " due to empty address pool";
					stats.minute_stats.connections_dropped ++;
				}
				else
				{
					// a new incoming TCP connection has been accepted

					if (mode == EPS::kClient)
					{
						// servers never call out, so if this is on a client, the only incoming connection can only be a mesh connection
						accepted_connection->is_mesh = true;
					}

					accepted_connection->accepted();
					this->validate_new_connection(accepted_connection);
				}

				if (not done)
				{
					this->start_async_accept();	// keep accepting new connections
				}
			} );

	return;
}


void EPS::Hull::async_write( const EPS::TConnPtr &connection, EPS::NetBuffer *nb)
{
	// This method returns *BEFORE* the NetBuffer has been written out to the socket.

	if (nb == nullptr)
	{
		LOG_ERROR << connection->name << "cannot write NULL buffer";
		return;
	}

	// if we don't have the length, update it from the header
	if (nb->length == 0 && nb->is_decrypted())
	{
		const EPS::CommHeader1 *header = nb->get_header();
		nb->length = header->size_of_packet();
		LOG_WARN << "buffer length was incorrectly set to zero; updating it to " << nb->length << " bytes (from header)";
	}

	if (nb->length == 0)
	{
		LOG_ERROR << connection->name << "invalid write of 0-byte buffer";
		delete nb;
		return;
	}

	// encrypt the data we need to write
	EPS::NetBuffer *encrypted_buffer = nb->get_encrypted();

	connection->pending_write_deque.push_back(encrypted_buffer);

	const size_t size_of_deque = connection->pending_write_deque.size();
	if (size_of_deque == 1)
	{
		// the buffer we just added to the deque is the only one there, meaning we can go ahead and start the write immediately
		this->async_write( connection );
	}
	else
	{
		// otherwise if we get here, then we're already in a pending write, so we'll have to wait our turn

		static time_t last_warning = 0;
		time_t now = std::time(nullptr);
		if (now > last_warning + 30)
		{
			if (size_of_deque > 75)
			{
				LOG_WARN << connection->name << "send deque now has " << size_of_deque << " packets waiting to be sent";
				last_warning = now;
			}
			else if (size_of_deque > 25)
			{
				LOG_DEBUG << connection->name << "send deque now has " << size_of_deque << " packets waiting to be sent";
				last_warning = now;
			}
		}
	}

	return;
}


void EPS::Hull::async_write( const EPS::TConnPtr &connection)
{
	if (connection->pending_write_deque.empty())
	{
		// nothing to write
		return;
	}

	/* NOTE:  The buffer MUST NOT be popped from the deque until it has been fully sent.  The existence of the deque
	 * is used to determine if there is a pending asynchronous write.  From the Boost asio documentation:
	 *
	 *		"The program must ensure that the stream performs no other write operations (such as async_write,
	 *		the stream's async_write_some function, or any other composed operations that perform writes) until
	 *		this operation completes."
	 */

	EPS::NetBuffer *nb = connection->pending_write_deque.front();
	EPS::NetBuffer *encrypted = nb->get_encrypted();

//	LOG_TRACE << connection->name << "preparing to asynchronously write " << encrypted->length << " encrypted bytes: " << EPS::hex_out(encrypted->data(), encrypted->length);

	// start the asynchronous write
	const boost::asio::const_buffers_1 write_buffer(encrypted->data(), encrypted->length);
	boost::asio::async_write(connection->socket, write_buffer,
			[=](const EC &ec, size_t bytes_written)
			{
				if (ec || bytes_written != encrypted->length)
				{
					LOG_ERROR << connection->name << "wrote " << bytes_written << " bytes instead of the expected " << encrypted->length << " bytes: " << ec.message();
					this->stats.minute_stats.packets_dropped ++;
					this->stats.minute_stats.bytes_dropped += encrypted->length - bytes_written;
					this->connection_failed(connection);
					delete nb;
					return;
				}

				if (connection->is_mesh)
				{
					this->stats.minute_stats.direct_packets_sent	++;
					this->stats.minute_stats.direct_bytes_sent		+= bytes_written;
				}
				else
				{
					this->stats.minute_stats.server_packets_sent	++;
					this->stats.minute_stats.server_bytes_sent		+= bytes_written;
				}

//				LOG_TRACE << connection->name << "successfully wrote " << bytes_written << " encrypted bytes asynchronously: " << EPS::hex_out(encrypted->data(), encrypted->length);

				// we're done with this NetBuffer
				connection->pending_write_deque.pop_front();
				delete nb;

				async_write(connection); // queue up any additional writes for this connection

				return;
			} );

	return;
}
