/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: common.cpp 2182 2017-03-26 08:30:05Z stephane $
 * See license.txt.
 */

#include "common.hpp"
#include "logging.hpp"
#include "comm.hpp"
#include "eps_conf.h"
#include <cctype>
#include <string>
#include <iomanip>
#include <algorithm>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include <random>
#include <cstdlib>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <time.h>
#include <stdio.h>
#include <openssl/rand.h>

#ifndef EPS_WINDOWS
	#include <net/ethernet.h>
#endif

#ifdef EPS_LINUX
#include <sys/sysinfo.h>
#endif

#ifdef EPS_FREEBSD
#include <sys/timespec.h>
#endif


static_assert(EPS::network_buffer_size > 1500, "network buffers need to be large enough to accommodate a typical MTU");


size_t EPS::number_of_pkcs_padding_bytes( const size_t &len )
{
	/** AES requires exactly 16-bytes blocks.  If a block is an exact multiple of 16 bytes, then another 16 bytes is
	 * added.  Otherwise, between 1 and 15 bytes of padding are added to get an exact multiple of 16.  This method
	 * returns the number of padding bytes needed to obtain a multiple of the AES block size.  @see PKCS #5 or PKCS #7
	 */

	static_assert(EPS::cipher_block_size == 16, "AES cipher block is expected to be 16 bytes in length.");

	size_t padding = 16 - (len % 16);
	if (padding == 0)
	{
		padding = 16;
	}

	return padding;
}


size_t EPS::length_including_pkcs_padding( const size_t &len )
{
	/// This method returns the full length, meaning our own size plus the number of padding bytes required.  @see PKCS #5 or PKCS #7

	const size_t padding = number_of_pkcs_padding_bytes(len);

	return len + padding;
}


std::string EPS::ipset_to_string(const IPv4Set &s)
{
	std::string str;

	for (const auto &ip : s)
	{
		str += (str.empty() ? "" : ", ");
		str += ip.to_string();
	}

	return str;
}


std::string EPS::mode_to_string(const EMode mode)
{
	return	mode == kClient ? "client" :
			mode == kServer ? "server" : "unknown";
}


void EPS::parse_network_and_mask(const std::string &str, IPv4 &network, IPv4 &netmask)
{
	boost::regex rx("^([0-9\\.]+)(\\/([0-9]{0,2}))?$");	// must parse strings such as 192.168.43.0/24
	boost::smatch what;

	// must match the *entire* string to be valid
	if (not boost::regex_match(str, what, rx))
	{
		THROW_INVALID_ARGUMENT(-1, "failed to parse network+netmask parameter \"" + str + "\".");
	}

	const std::string network_str = what[1];
	const std::string netmask_str = what[3];

	EC ec;
	network = IPv4::from_string(network_str, ec);
	if (ec)
	{
		THROW_INVALID_ARGUMENT(-1, "failed to parse network argument \"" + network_str + "\": " + ec.message());
	}

	if (netmask_str != "24" && not netmask_str.empty())
	{
		/// @todo only 24-bit "class C" networks are currently supported
		THROW_INVALID_ARGUMENT(-1, "netmask must be /24 (other options not yet supported)");
	}
	netmask = IPv4::from_string("255.255.255.0", ec);

	// check to see if the network and the netmask make sense
	//
	// for example:		network 192.168.43.0
	//					netmask 255.255.0.0
	//					AND ==  192.168.0.0		-> result does not match the original network
	//
	if ((network.to_ulong() & netmask.to_ulong()) != network.to_ulong())
	{
		THROW_INVALID_ARGUMENT(-1, "netmask " + netmask.to_string() + " cannot be used with the network " + network.to_string());
	}

	LOG_INFO << "parse " << str << ": network is " << network.to_string() << ", netmask is " << netmask.to_string();

	return;
}


std::string EPS::proto_to_string(const uint16_t proto_in_host_order)
{
	return	proto_in_host_order == ETHERTYPE_IP		?	"ipv4"	:
			proto_in_host_order == ETHERTYPE_IPV6	?	"ipv6"	:
			proto_in_host_order == ETHERTYPE_ARP	?	"arp"	:
														"unknown (" + std::to_string(proto_in_host_order) + ")";
}


std::string EPS::proto_to_string(const EPS::CommData1 *	cd1)
{
	/* tap packets look like this:
	 *
	 *	- byte[0-1]:	flags (network byte order)
	 *	- byte[2-3]:	proto (network byte order)
	 *  - byte[4...]:	ethernet frame
	 *
	 * ...except that we're now setting the IFF_NO_PI flag when creating the tap interface, so the flags and proto
	 * fields are no longer being sent.  Now, the packets look like this:
	 *
	 *  - byte[0...]:	ethernet frame
	 *	- byte[...?]:	IP, etc...
	 */

	/* If __NOT__ using IFF_NO_PI then this next line works:
	 *
	const uint16_t *proto = reinterpret_cast<const uint16_t*>(&cd1->data[2]);
	 *
	 */

	/* Otherwise, now that we no longer have those initial 4 header bytes, we'll go peek ourselves at the 2-byte
	 * ethertype value within the ethernet frame.
	 */
	const uint16_t *proto = reinterpret_cast<const uint16_t*>(&cd1->data[12]);

	return proto_to_string(htons(*proto));
}


std::string EPS::format(const long double &value, std::string &postfix)
{
	// network traffic is always base 1000, never base 1024

	const char	modifier[]	= " KMGTPEZYB";
	long double	divisor		= 1000.0;

	size_t idx = 0;
	while (value > divisor/2)
	{
		idx ++;
		divisor *= 1000.0;
	}

	std::stringstream ss;
	ss << std::fixed << std::setprecision(2) << value/(divisor/1000.0);
	if (idx)
	{
		postfix = modifier[idx];
	}
	else
	{
		postfix.clear();
	}

	return ss.str();
}


std::string EPS::format_bytes(const uint64_t &value)
{
	const long double ld = (long double)value;
	std::string postfix;
	std::string str = format(ld, postfix);

	str += " " + postfix + "Byte";
	if (value != 1)
	{
		str += "s";
	}

	return str;
}


std::string EPS::format_bps(const uint64_t &value, int seconds)
{
	long double ld = value / (long double)seconds;
	std::string postfix;
	std::string str = format(ld, postfix);

	str += " " + postfix + "Bps";

	if (value != 0 )
	{
		const long double bps = (long double)seconds / 8.0;
		ld = value / bps;
		str += " [" + format(ld, postfix);
		str += " " + postfix + "bps]";
	}

	return str;
}


int EPS::get_random(const int min_val, const int max_val)
{
	if (min_val > max_val)
	{
		THROW_INVALID_ARGUMENT(-1, "min value should be smaller than max value (min=" + std::to_string(min_val) + ", max=" + std::to_string(max_val) + ")");
	}

	if (min_val == max_val)
	{
		// not much room for randomness...
		return min_val;
	}

	static std::default_random_engine generator;
	static bool seed_needed = true;
	if (seed_needed)
	{
		uint32_t seed;	// 4 bytes long, don't initialize (on purpose; see RAND_bytes() for details)
		uint8_t *ptr = reinterpret_cast<uint8_t*>(&seed);
		RAND_bytes(ptr, sizeof(seed));	// use OpenSSL to generate a decent seed value
		srand(seed);
		generator.seed(rand());
		seed_needed = false;
	}

	std::uniform_int_distribution<int> distribution(min_val, max_val);

	return distribution(generator);
}


std::string EPS::read_file(const std::string &filename)
{
	LOG_TRACE << "reading " << filename;

	std::string line;
	std::stringstream ss;
	std::ifstream ifs(filename);
	size_t line_count = 0;

	while (std::getline(ifs, line))
	{
		line_count ++;
		ss << line << std::endl;
	}

	const std::string str = ss.str();

	LOG_TRACE << "read " << line_count << " lines (" << str.size() << " bytes) from " << filename;

	return ss.str();
}


void EPS::write_file(const std::string &filename, const std::string &text)
{
	const size_t line_count = std::count(text.begin(), text.end(), '\n');
	LOG_TRACE << "writing " << line_count << " lines (" << text.size() << " bytes) to " << filename;

	const std::string filename_tmp = filename + "_" + std::to_string(getpid());

	std::ofstream ofs(filename_tmp);
	ofs << text;
	ofs.close();

	const int rc = rename(filename_tmp.c_str(), filename.c_str());
	if (rc)
	{
		LOG_ERROR << "error renaming " << filename_tmp << " to " << filename << " (rc=" << rc << ")";
	}

	return;
}


std::string EPS::hex_out(const uint8_t *data, const size_t len, const int number_of_rows_to_display)
{
	/* Format the data into hex/text output like the following, usually for debug logging:
	 *
	 *	number of bytes: 368
	 *	   0: 8b 3c e8 bf c6 cf d2 81 a5 f0 a1 ab 11 b4 56 8f ee 91 d8 9e b9 ab 62 94 bc 74 77 20 a0 45 50 ad - .<............V.......b..tw .EP.
	 *	  20: 28 db 3c e7 89 9c 30 c0 21 9a ad 96 6f da e5 01 73 53 b0 ef bf 08 14 12 cb 92 1f e9 0d ac e8 06 - (.<...0.!...o...sS..............
	 *	  40: 0c d3 51 b6 8b 12 6c c0 a6 3c 4e f7 67 11 7f 6a fe 2e 88 8d 76 9f 2f 64 a0 aa 6e f6 95 52 84 32 - ..Q...l..<N.g..j....v./d..n..R.2
	 *	...
	 */

	const size_t width = 32;	// number of "characters" to output per line (normally 16 or 32)

	std::stringstream ss;
	ss << "number of bytes: " << len;

	size_t skip_start	= 1; // initially set to a value we'd never see at the start of a row
	size_t skip_end		= 0;
	size_t skip_rows	= 0;
	size_t skip_bytes	= 0;

	/** If number_of_rows_to_display is set to 1 or less, then all rows will be shown without skipping.  Otherwise,
	 * the middle section of large buffers will be skipped to reduce the number of lines logged, keeping both the
	 * start and end of the buffer intact.
	 */
	if (number_of_rows_to_display > 1 && len > (width * (number_of_rows_to_display + 2)))
	{
		// if we have too many rows, then skip the middle ones so we just output the first few and last few lines
		skip_start	= 0		+ width * number_of_rows_to_display / 2;
		skip_end	= len	- width * number_of_rows_to_display / 2;	// this is just an approximation (we actually want a multiple of "width")
		skip_rows	= (skip_end - skip_start) / width;
		skip_end	= skip_start + skip_rows * width;	// now we have the exact end offset
		skip_bytes	= skip_end - skip_start;
	}

	size_t offset = 0;
	while (offset < len)
	{
		if (offset == skip_start)
		{
			ss << std::endl << "  --: ...skipping " << skip_rows << " rows (" << skip_bytes << " bytes)...";
			offset = skip_end;
		}

		std::stringstream line;
		std::stringstream text;

		ss << std::endl << std::right << std::setw(4) << std::hex << offset << ":" << std::dec;
		line << std::setfill('0') << std::hex;

		for (size_t count = 0; count < width && offset < len; count ++, offset ++)
		{
			const uint8_t c = data[offset];
			line << " " << std::setw(2) << (int)c;
			text << char(c >= 32 && c <= 126 ? c : '.');
		}

		ss	<< std::left << std::setw(width*3) << line.str()	// the hex values of each character
			<< " - " << text.str();								// the text representation (if available)
	}

	return ss.str();
}


std::string EPS::get_utc_timestamp(time_t tt)
{
	if (tt == 0)
	{
		tt = std::time(nullptr);
	}

	const tm *gmt = std::gmtime(&tt);
	char buffer[50];
	std::strftime(buffer, sizeof(buffer), "%Y-%m-%d %H:%M:%S UTC", gmt);

	return buffer;
}


std::string EPS::to_hex_string(const uint8_t *data, const size_t len)
{
	std::stringstream ss;
	for (size_t idx=0; idx < len; idx++)
	{
		ss << std::hex << std::setw(2) << std::setfill('0') << (int)data[idx];
	}

	return ss.str();
}


void EPS::from_hex_string(uint8_t *data, const size_t len, std::string str)
{
	const boost::regex rx(
		"^"					// start of line
		"\\s*"				// leading whitespace
		"([A-Fa-f0-9]+)"	// hex value
		"\\s*"				// trailing whistespace
		"$"					// end of line
		);

	boost::smatch what;
	if (not boost::regex_match(str, what, rx))
	{
		THROW_INVALID_ARGUMENT(-1, "Not a valid " + std::to_string(len) + "-byte hex string: " + str);
	}
	str = what[1].str();

	if (str.size() != len * 2)
	{
		THROW_INVALID_ARGUMENT(-1, "Expected " + std::to_string(len*2) + " hex bytes, but found " + std::to_string(str.size()) + ": \"" + str + "\"");
	}

	for (size_t idx = 0; idx < len; idx++)
	{
		std::string hex = str.substr(idx*2, 2);
		const unsigned long ul = strtoul(hex.c_str(), nullptr, 16);
		data[idx] = (uint8_t)ul;
	}

	return;
}


std::string EPS::get_eps_conf_template(void)
{
	/// The contents of the template eps.conf file are generated from eps.ini using xxd.  @see eps_conf.ini  @see CMakeLists.txt
	return std::string((const char *)eps_conf_template, (size_t)eps_conf_template_len);
}


std::string EPS::get_os_name(void)
{
	static std::string name;
	if (name.empty())
	{
		#if	! defined( EPS_LINUX	) && \
			! defined( EPS_FREEBSD	) && \
			! defined( EPS_WINDOWS	)
			// I know the following conditions in this function works on operating systems like Ubuntu, CentOs,
			// Fedora, FreeBSD, and Windows.  If adding a new system type, then see if we need to add new conditions.
			#error "See if the following code will get a valid name for this new OS."
		#endif

		#ifdef EPS_WINDOWS

			/** On Windows, the OS name can be found in "HKEY_LOCAL_MACHINE\Software\Microsoft\Windows NT\CurrentVersion\ProductName",
			 * which should contain a short string such as "Windows 7 Professional".
			 */
			name = "Windows";
			HKEY hkey;
			RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows NT\\CurrentVersion", 0, KEY_READ , &hkey);
			char buffer[512];
			DWORD buffer_size = sizeof(buffer);
			const ULONG error = RegQueryValueEx(hkey, "ProductName", 0, NULL, (LPBYTE)buffer, &buffer_size);
			if (error == ERROR_SUCCESS)
			{
				name = buffer;
			}
			RegCloseKey(hkey);

		#else

			std::string text = EPS::read_file("/etc/os-release");
			/** From the `/etc/os-release` text file, try and extract @p PRETTY_NAME.  For example:
			 *
			 * 		VERSION_ID="7"
			 * 		PRETTY_NAME="CentOS Linux 7 (Core)"
			 * 		ANSI_COLOR="0;31"
			 *
			 * or:
			 *
			 * 		ID_LIKE=debian
			 * 		PRETTY_NAME="Ubuntu 14.04.1 LTS"
			 * 		VERSION_ID="14.04"
			 *
			 * or:
			 *
			 * 		VERSION_ID=20
			 * 		PRETTY_NAME="Fedora 20 (Heisenbug)"
			 * 		ANSI_COLOR="0;34"
			 */

			boost::regex rx("PRETTY_NAME=\"([^\"]+)\"");
			boost::smatch what;
			if (boost::regex_search(text, what, rx))
			{
				name = what[1].str();
			}

			/// If os-release doesn't yield a usable name, see if `/etc/issue` has something we can use.
			if (name.empty())
			{
				/* for example:
				 *
				 *		Ubuntu 14.04.1 LTS \n \l
				 */
				text = EPS::read_file("/etc/issue");
				rx = boost::regex("^([^\\\r\n]+)");
				if (boost::regex_search(text, what, rx))
				{
					name = what[1].str();
				}
			}

			if (name.empty())
			{
				/// Otherwise, the name used will be the output from running `uname -sr`, such as "Linux 3.13.0-37-generic".

				FILE *file = popen("uname -sr", "r");
				if (file)
				{
					char buffer[100];
					const char *ptr = fgets(buffer, sizeof(buffer), file);
					if (ptr)
					{
						name = ptr;
					}
					pclose(file);
				}
			}
		#endif

		// remove leading and trailing whitespace
		boost::algorithm::trim(name);
	}

	LOG_DEBUG << "operating system: \"" << name << "\"";

	return name;
}


std::string EPS::get_approximate_time(const time_t timestamp)
{
	/** Figure out how long ago was "timestamp".  Return an approximate text string, not a precise value.  As the
	 * timestamp gets older, the precision is eroded.  For example, valid strings returned could be "3 minutes ago"
	 * or "3 months ago".
	 */

	const time_t now = std::time(nullptr);

	if (timestamp > now)
	{
		// timestamp is in the future?
		return "unknown";
	}

	const double seconds	= 1.0;
	const double minutes	= 60.0		* seconds;
	const double hours		= 60.0		* minutes;
	const double days		= 24.0		* hours;
	const double weeks		= 7.0		* days;
	const double years		= 365.2425	* days;		// Gregorian calendar: 365 + 97/400 = 365.2425
	const double months		= years	/ 12.0;
	const time_t duration	= now	- timestamp;

	if (duration <= 2.0 * minutes)
	{
		std::string str = std::to_string(duration) + " second";
		if (duration != 1)
		{
			str += "s";
		}
		return str;
	}

	if (duration <= 2.0 * hours)
	{
		const int value = (int) (duration / minutes);
		return std::to_string(value) + " minutes";
	}

	// once we get past seconds and minutes, use a single decimal point in the values
	std::stringstream ss;
	ss << std::fixed << std::setprecision(1);

	if (duration <= 2.0 * days)
	{
		ss << duration / hours << " hours";
		return ss.str();
	}

	if (duration <= 2.0 * weeks)
	{
		ss << duration / days << " days";
		return ss.str();
	}

	if (duration <= 2.0 * months)
	{
		ss << duration / weeks << " weeks";
		return ss.str();
	}

	if (duration <= 2.0 * years)
	{
		ss << duration / months << " months";
		return ss.str();
	}

	if (duration > 5.0 * years)
	{
		// unlikely...timestamp is from 5 years ago?
		return "more than 5 years";
	}

	ss << duration / years << " years";
	return ss.str();
}


time_t EPS::get_boot_timestamp(void)
{
	// timestamp representing when the system was last rebooted

	const time_t	now					= std::time(nullptr);
	time_t			time_of_last_reboot	= now;

	#ifdef EPS_LINUX

		struct sysinfo info;
		sysinfo(&info);

		// but we don't want to know how long it has been, we want the *timestamp* when it rebooted
		time_of_last_reboot = now - info.uptime;

	#elif EPS_FREEBSD

		timespec tp;
		clock_gettime(CLOCK_UPTIME, &tp);
		time_of_last_reboot = now - tp.tv_sec;

	#elif EPS_WINDOWS

		time_of_last_reboot = now - (GetTickCount64() / 1000);

	#else
		#error "Insert code here for your operating system."
	#endif

	return time_of_last_reboot;
}


std::string &EPS::trim( std::string &str )
{
	// trim the back first to reduce the length of the string
	size_t len = str.size();
	while (len > 0)
	{
		const size_t idx = len - 1;

		if ( std::isspace(str[idx]) )
		{
			str.erase( idx );
		}
		else
		{
			break;
		}

		len --;
	}

	// now trim the front of the string
	while (len > 0)
	{
		if ( std::isspace(str[0]) )
		{
			str.erase( 0, 1 );
		}
		else
		{
			break;
		}

		len --;
	}

	return str;
}


std::string trim( const std::string &str )
{
	std::string new_str( str );
	trim(new_str);
	return new_str;
}
